@@ -1,8 +1,8 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
 
-import data_dispenser.sources
+from data_dispenser.sources import Source
 
 __author__ = 'Catherine Devlin'
 __email__ = 'catherine.devlin@gmail.com'
-__version__ = '0.1.1'
+__version__ = '0.2.0'
