@@ -6,7 +6,7 @@ a source of row-like data, acts as a generator returning
 OrderedDicts for each row.
 """
 from collections import OrderedDict
-from io import StringIO
+from io import StringIO, BytesIO
 import csv
 import doctest
 import glob
@@ -16,6 +16,7 @@ import logging
 import os.path
 import pickle
 import pprint
+import sys
 import xml.etree.ElementTree as et
 try:
     import yaml
@@ -27,6 +28,16 @@ try:
 except ImportError:
     logging.info("Could not import Collection from pymongo; is it installed?")
     MongoCollection = None.__class__
+try:
+    import requests
+except ImportError:
+    logging.info("Could not import ``requests``, will not load from URLs")
+    requests = None
+try:
+    import xlrd
+except ImportError:
+    logging.info("Could not import ``xlrd``, will not load from .xls")
+    xlrd = None
 
 if yaml:
     def ordered_yaml_load(stream, Loader=yaml.Loader,
@@ -93,9 +104,25 @@ def _first_list_in(element):
 def _ensure_rows(result):
     """data_dispenser is for rowlike sources.  If the data source
        evaluates to a single dict instead of a listlike object,
-       return a list (that contains the dict as its sole row). """
+       transform a dict of dicts into a list of dicts, or 
+       return a list (that contains the dict as its sole row).
+       >>> pprint.pprint(_ensure_rows({"a": 1, "b": 2}))
+       [{'a': 1, 'b': 2}]
+       >>> pprint.pprint(_ensure_rows({"a": {"a1": 1, "a2": 2}, "b": {"b1": 1, "b2": 2}}))
+       [{'a1': 1, 'a2': 2, 'name_': 'a'}, {'b1': 1, 'b2': 2, 'name_': 'b'}]
+       
+       otherwise just don't mess with it
+       >>> pprint.pprint(_ensure_rows([{"a1": 1, "a2": 2}, {"b1": 1, "b2": 2}]))
+       [{'a1': 1, 'a2': 2}, {'b1': 1, 'b2': 2}]
+    """
     if isinstance(result, dict):
-        result = [result, ]
+        if not result:
+            result = []
+        # if it's a dict of dicts, convert to a list of dicts
+        if not [s for s in result.values() if not hasattr(s, 'keys')]:
+            result = [dict(name_=k, **result[k]) for k in result]
+        else:
+            result = [result, ]
     return result
 
 
@@ -125,11 +152,12 @@ def _eval_file_obj(target):
     else:
         yield result
 
-def _eval_csv(target):
+def _eval_csv(target, fieldnames=None):
     """
     Yields OrderedDicts from a CSV string
     """
-    reader = csv.DictReader(target)
+    # TODO: automatic FieldA, FieldB, etc
+    reader = csv.DictReader(target, fieldnames=fieldnames)
     for row in reader:
         yield OrderedDict((k, row[k]) for k in reader.fieldnames)
 
@@ -142,6 +170,14 @@ def _open(filename):
     input_source = open(filename, file_mode)
     return input_source
 
+class NamedIter(object):
+    "Hack to let us assign attributes to an iterator"
+
+    def __init__(self, unnamed_iterator):
+        self.__iter__ = unnamed_iterator.__iter__
+        self.__next__ = unnamed_iterator.__next__
+
+
 class Source(object):
     """
     A universal data generator that returns one "row" at
@@ -177,6 +213,7 @@ class Source(object):
     eval_funcs_by_ext = {'.py': [_eval_file_obj, ],
                          '.json': [json_loader, ],
                          '.yaml': [ordered_yaml_load, ],
+                         '.yml': [ordered_yaml_load, ],
                          '.csv': [_eval_csv, ],
                          '.xml': [_eval_xml, ],
                          '.pickle': [pickle_loader, ],
@@ -190,7 +227,8 @@ class Source(object):
     table_count = 0
 
     def _source_is_generator(self, src):
-        self.table_name = src.name
+        if hasattr(src, 'name'):
+            self.table_name = src.name
         self.generator = src
         return
 
@@ -241,18 +279,61 @@ class Source(object):
         input_source = _open(src)
         self._deserialize(input_source)
 
-    def _source_is_glob(self, sources):
+    def _multiple_sources(self, sources):
         subsources = [Source(s, limit=self.limit) for s in sources]
         self.limit = None  # impose limit only on the subsources
         self.generator = itertools.chain.from_iterable(subsources)
 
+    def _source_is_url(self, src):
+        if not requests:
+            raise ImportError('must ``pip install requests to read from web``')
+        (core_url, ext) = os.path.splitext(src)
+        ext = ext.lower()
+        response = requests.get(src)
+        if src.endswith('.xls'):
+            return self._source_is_excel(response.content)
+        self.deserializers = self.eval_funcs_by_ext[ext or '*']
+        if ext == '.pickle':
+            self._deserialize(BytesIO(response.content))
+        else:
+            content = response.content.decode(response.encoding or response.apparent_encoding)
+            self._deserialize(StringIO(content))
+
     def _source_is_open_file(self, src):
         if hasattr(src, 'name'):
             self.table_name = src.name
         self.deserializers = self.eval_funcs_by_ext['*']
         self._deserialize(src)
 
-    def __init__(self, src, limit=None):
+    def _source_is_excel(self, spreadsheet):
+        if not xlrd:
+            raise ImportError('must ``pip install xlrd``')
+        if len(spreadsheet) < 84 and spreadsheet.endswith('xls'):
+            workbook = xlrd.open_workbook(spreadsheet)
+            name = spreadsheet
+        else:
+            workbook = xlrd.open_workbook(file_contents=spreadsheet)
+            name = "excel"
+        generators = []
+        for sheet in workbook.sheets():
+            headings = ["Col%d" % c for c in range(1, sheet.ncols + 1)]
+            data = []
+            for row_n in range(sheet.nrows):
+                row_has_data = max(bool(v) for v in sheet.row_values(row_n))
+                if row_has_data:
+                    headings = [heading if heading else default_heading
+                                for (heading, default_heading) 
+                                in itertools.zip_longest(sheet.row_values(row_n), headings)]
+                    row_n += 1
+                    break
+            data = [OrderedDict(zip(headings, sheet.row_values(r)))
+                                for r in range(row_n,sheet.nrows)]
+            generator = NamedIter(iter(data))
+            generator.name = "%s-%s" % (name, sheet.name)
+            generators.append(generator)
+        self._multiple_sources(generators)
+
+    def __init__(self, src, limit=None, fieldnames=None):
         self.counter = 0
         self.limit = limit
         self.deserializers = []
@@ -261,6 +342,10 @@ class Source(object):
         if isinstance(src, MongoCollection):
             self._source_is_mongo(src)
             return
+        if hasattr(src, 'startswith') and (
+                src.startswith("http://") or src.startswith("https://")):
+            self._source_is_url(src)
+            return
         if hasattr(src, 'read'):    # open file
             self._source_is_open_file(src)
             return
@@ -269,20 +354,27 @@ class Source(object):
             return
         try:
             if os.path.isfile(src):
-                self._source_is_path(src)
+                if src.endswith('.xls'):
+                    self._source_is_excel(src)
+                else:
+                    self._source_is_path(src)
                 return
         except TypeError:
             pass
+        if hasattr(src, 'startswith') and src.startswith('http'):
+            self._source_is_url(src)
         try:
             data = eval(src)
+            if not hasattr(data, '__next__'):
+                data = iter(data)
             self._source_is_generator(data)
             return
         except:
             pass
         try:
-            sources = glob.glob(src)
+            sources = sorted(glob.glob(src))
             if sources:
-                self._source_is_glob(sources)
+                self._multiple_sources(sources)
                 return
         except:
             pass
@@ -311,4 +403,8 @@ class Source(object):
 
 
 if __name__ == '__main__':
-    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)
+    if len(sys.argv) > 1:
+        for target in sys.argv[1:]:
+            pprint.pprint(list(Source(target)))
+    else:
+        doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)
